#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdint.h>

typedef uint32_t	u32;
typedef uint16_t	u16;
typedef uint8_t		u8;
typedef int32_t		s32;
typedef int16_t		s16;
typedef int8_t		s8;

#define DACP OCR0A
#define DACN OCR0B

__attribute__((naked)) int main(void)
{
//	CLKPR = 0b10000000; // 16 MHz from the high freq PLL clock
//	CLKPR = 0b00000000;
	// Configure ports

	DDRB  = 0b11110111; // OC0AB are L/R PWM output. Drive unused pins low (helps noise).
	PORTB = 0b00000000; // No activate pull-ups
//	DDRB  = 0b00000011;

	// ADC
	ADMUX  = 0b00010011; // PB3, AVcc, 10 bit, 2.56VRef
	ADCSRA = 0b11001101; // CPU/32 sampling rate; Sampling freq: 16 MHz / 32 / 13 = 39 kHz

	// Power reduction
	PRR   = 0b00001010; // Clock-gate USI and Counter1.
	MCUCR = 0b00100000;


	// PWM
	TCCR0A = 0b10100011; // Fast PWM
	TCCR0B = 0b00000001; // Increment TCNT0 at CPU; Sampling freq: 16MHz / 256 = 64 kHz (= MAX)
	//TIMSK  = 0b10; // Interrupt on overflow

	sei();

	while(1) {
		sleep_cpu();
	}
}

#if 0

inline s16 min(s16 a, s16 b) {
	return (a<b)?a:b;
}

inline s16 max(s16 a, s16 b) {
	return (a>b)?a:b;
}

// 0% duty cycle at idle. Pulses too narrow at low volume.
ISR(ADC_vect) {
	static u8 last_error;

	u16 adc		= ADC;

	s16 val		= (s16)(adc >> 1) - 256;

	u8  error	= adc & 1;
	u8  dither	= error & last_error;

	last_error	= error - dither;

	val += dither;
	val  = min(max(val, -255),255);

	DACN = max(val, 0);
	DACP = max(-val, 0);
}

#else
// 50% duty cycle at idle. Slightly less efficient, but better sounding.
ISR(ADC_vect) {
//ISR(TIMER0_OVF_vect) {
	static u8 last_error;

	u16 adc		= ADC;
	s8 val1		= (u8)(adc >> 2) - 128;
	u8 val2		= (u8)(adc & 0b10) >> 1;
	u8 error	= adc & 1;
	u8 dither	= error & last_error;
	last_error	= error - dither;

	u8 pos = 128 + val1 + val2;
	u8 neg = 128 - val1 - dither;

	DACP = pos;
	DACN = neg;

	if(val1 == 127)
		DACP = 255;

	if(val1 == -128)
		DACN = 255;

}

#endif


